"""Coordinate chart.

Creates a grid of tiles with consistently spaced x/y positions. Provided data is plotted as a marker at each position
unless None. This can be used to create all sorts of visualizations that have spatially related data, such as calendars

"""

import calendar
import cmath
import math
from itertools import chain

import numpy as np
import pandas as pd
import plotly.graph_objects as go

from .utils_fig import CustomChart, check_raw_data

# PLANNED: subplots for multiple years of calendar charts (Subplot title is year)


def calculate_grid(grid_dims, corners, width, height):
    """Calculate the grid x and y coordinates.

    Args:
        grid_dims: tuple of the number of tiles in grid. In format `(row, column)`
        corners: dictionary with keys `(x, y)` containing lists of the four exterior corner coordinates
        width: float width in pixels
        height: float height in pixels

    Returns:
        dict: with keys `(x, y)` with lists of lists containing float values

    """
    grid = {'x': [], 'y': []}
    for r_idx in range(grid_dims[0]):
        y_offset = height * (grid_dims[0] - r_idx)
        y_grid = [y_offset - _y for _y in corners['y']]
        for c_idx in range(grid_dims[1]):
            x_offset = width * c_idx
            grid['x'].extend([x_offset + _x for _x in corners['x']])
            grid['y'].extend(y_grid)
    return grid


def calculate_border(grid_dims, width, height):
    """Calculate each line in all borders.

    Args:
        grid_dims: tuple of the number of tiles in grid. In format `(row, column)`
        width: float width in pixels
        height: float height in pixels

    Returns:
        list: containing dictionaries keys `(x, y)` and values for the two points for each line in grid

    """
    return [
        {
            'x': [c_idx * width] * 2,
            'y': [0, height * grid_dims[0]],
        } for c_idx in range(grid_dims[1] + 1)
    ] + [
        {
            'x': [0, width * grid_dims[1]],
            'y': [r_idx * height] * 2,
        } for r_idx in range(grid_dims[0] + 1)
    ]


class CoordinateChart(CustomChart):  # noqa: H601
    """Coordinate Chart."""

    border_opacity: float = 0.2
    """Border opacity for grid. Value must be in [0-1] where 0 is none. Default is 0.2."""

    border_line = None
    """Dictionary passed to plotly `line`. Used to set thickness, color, dash style, etc. Default is None."""

    marker_kwargs = None
    """Marker keyword arguments used in `create_marker()`. Default is None."""

    # Private states for managing coordinate chart dimensions
    _grid: dict
    _borders: list

    def __init__(self, *, title, grid_dims, corners, titles=None, layout_overrides=()):
        """Initialize Coordinate Chart and store parameters as data members.

        Args:
            title: String title for chart (can be an empty string for blank)
            grid_dims: tuple of the number of tiles in grid. In format `(row, column)`
            corners: dictionary with keys `(x, y)` containing lists of the four corner coordinates
            titles: list of strings that will appear in each tile. Default is None for no titles
            layout_overrides: Custom parameters in format [ParentKey, SubKey, Value] to customize 'go.layout'

        """
        # Initialize base method. Sets xlabel and ylabel to empty strings because coordinate chart use the x/y
        #   axis for arranging points. Data is displayed by color or size
        super().__init__(title=title, xlabel='', ylabel='', layout_overrides=layout_overrides)

        # Initialize chart parameters
        self.calculate_layout(grid_dims, corners, titles)

    def calculate_layout(self, grid_dims, corners, titles):
        """Calculate coordinate chart layout. Called by __init__, but can be called later to update the chart.

        Args:
            grid_dims: tuple of the number of tiles in grid. In format `(row, column)`
            corners: dictionary with keys `(x, y)` containing lists of the four corner coordinates
            titles: list of strings that will appear in each tile. Default is None for no titles

        """
        # Calculate exterior height and width of grid
        width = float(np.max(corners['x']) + np.min(corners['x']))
        height = float(np.max(corners['y']) + np.min(corners['y']))

        # Set grid and border coordinates for traces
        self._grid = calculate_grid(grid_dims, corners, width, height)
        self._borders = calculate_border(grid_dims, width, height)

        # Add titles to annotations if provided
        if titles is None:
            self.annotations = []
        else:
            v_offset = np.min(corners['y']) * 0.4
            self.annotations = [
                go.layout.Annotation(
                    ax=0, ay=0,
                    x=(idx % grid_dims[1] + 0.5) * width,  # noqa: S001
                    y=(grid_dims[0] - int(idx / grid_dims[1]) % grid_dims[0]) * height - v_offset,
                    text=title,
                )
                for idx, title in enumerate(titles) if title is not None
            ]

    def create_traces(self, df_raw):
        """Return traces for plotly chart.

        Args:
            df_raw: pandas dataframe with at minimum the column `values: str`

        Returns:
            list: Dash chart traces

        """
        # Check that the raw data frame is properly formatted
        check_raw_data(df_raw, min_keys=['values'])

        # Merge x/y grid data with values. Temporarily extend values with None, then drop those rows
        values = df_raw['values'].to_list()
        values.extend([None] * (len(self._grid['x']) - len(values)))
        df_grid = pd.DataFrame(
            data={
                'values': values,
                'x': self._grid['x'],
                'y': self._grid['y'],
            },
        ).dropna()

        return [
            go.Scatter(
                hoverinfo='none',
                line=self.border_line or {'color': 'black'},
                mode='lines',
                opacity=self.border_opacity,
                showlegend=False,
                x=border['x'],
                y=border['y'],
            ) for border in self._borders
        ] + [
            go.Scatter(
                hoverinfo='text',
                mode='markers',
                showlegend=False,
                text=df_grid['values'],
                x=df_grid['x'],
                y=df_grid['y'],
                marker=self.create_marker(df_grid, **(self.marker_kwargs or {})),
            ),
        ]

    def create_marker(self, df_grid, colorscale='Viridis', size=16, symbol='circle'):
        """Return a dictionary for the scatter plot.

        See: https://plot.ly/python/colorscales/ (Named colorscales: Reds, Bluered, Jet, Viridis, Cividis, etc.)

        Args:
            df_grid: pandas dataframe with at minimum the column `values: str`, `x: float`, `y: float`
            colorscale: plotly colorscale, see doc link above. Default is 'Viridis'
            size: integer marker size
            symbol: marker symbol (square, circle, circle-open, x, etc.)

        Returns:
            dict: the chart marker shape, symbol, color, etc.

        """
        return {
            'color': df_grid['values'],
            'colorscale': colorscale,
            'showscale': True,
            'size': size,
            'symbol': symbol,
        }

    def create_layout(self):
        """Extend the standard layout.

        Returns:
            dict: layout for Dash figure

        """
        layout = super().create_layout()
        for axis in ['xaxis', 'yaxis']:
            layout[axis]['showgrid'] = False
            layout[axis]['showticklabels'] = False
            layout[axis]['zeroline'] = False
        layout['yaxis']['scaleanchor'] = 'x'
        layout['yaxis']['scaleratio'] = 1
        return layout


# ==============================================================================
# Standard Coordinate Grids


class GridClass:
    """Base class for specifying a grid for the Coordinate chart."""

    marker_kwargs = None
    """Marker keyword arguments. Default is None."""

    def __init__(self, grid_dims, titles):
        """Initialize the coordinates.

        Args:
            grid_dims: tuple of the number of tiles in grid. In format `(row, column)`
            titles: list of titles to place in each grid element

        """
        self.grid_dims = grid_dims
        self.titles = titles


class CircleGrid(GridClass):  # noqa: H601
    """Grid of circular coordinates."""

    marker_kwargs = {'size': 10}
    """Marker keyword arguments. Default is `{'size': 10}`"""

    def __init__(self, grid_dims, titles=None):
        """Initialize the coordinates.

        Args:
            grid_dims: tuple of the number of tiles in grid. In format `(row, column)`
            titles: list of titles to place in each grid element. Default is None

        """
        if titles is None:
            titles = []
            for idx in range(grid_dims[0] * grid_dims[1]):
                x_coord = int(idx / grid_dims[1]) + 1
                y_coord = idx % (grid_dims[0] + 1) + 1  # noqa: S001
                titles.append(f'Subtitle for ({x_coord}, {y_coord})')

        super().__init__(grid_dims=grid_dims, titles=titles)

        # Calculate four corners
        opp = 0.5 * math.cos(cmath.pi / 4)
        adj = 0.5 * math.sin(cmath.pi / 4)
        self.corners = {
            'x': [0.5, 1 - adj, 1.0, 1 + adj, 1.5, 1 + adj, 1.0, 1 - adj],
            'y': [1.0, 1 - opp, 0.5, 1 - opp, 1.0, 1 + opp, 1.5, 1 + opp],
        }


def calculate_calendar_grid_corners(margin, days_in_week=7, max_weeks_in_month=6):
    """Calculate the four exterior corner coordinates of a calendar coordinate grid.

    Args:
        margin: float spacing between tiles
        days_in_week: number of days in week. Default is 7
        max_weeks_in_month: max number of weeks in a month. Default is 6

    Returns:
        list: dictionary with keys `(x, y)` containing lists of the four exterior corner coordinates

    """
    y_indices = [[idx] * days_in_week for idx in range(max_weeks_in_month)]
    return {
        'x': np.add([*range(days_in_week)] * max_weeks_in_month, margin),
        'y': np.add([*chain.from_iterable(y_indices)], margin),
    }


class YearGrid(GridClass):  # noqa: H601
    """Coordinates of days within a grid of months in one year."""

    marker_kwargs = {'size': 10, 'symbol': 'square'}
    """Marker keyword arguments. Default is `{'size': 10, 'symbol': 'square'}`"""

    def __init__(self, grid_dims=(4, 3), titles=None):
        """Initialize the coordinates.

        Args:
            grid_dims: tuple of the number of tiles in grid. In format `(row, column)`. Default is (4, 3)
            titles: list of titles to place in each grid element. Default is None

        Raises:
            RuntimeError: if error in the grid dimensions

        """
        if grid_dims[0] * grid_dims[1] != 12:  # pragma: no cover
            raise RuntimeError('Calendar must show all 12 months Expected (12,1), (6,2), (4,3), (1,12), etc.')
        if titles is None:
            titles = calendar.month_name[1:]

        super().__init__(grid_dims=grid_dims, titles=titles)

        # Calculate four corners
        self.corners = calculate_calendar_grid_corners(margin=2)

    def format_data(self, month_lists, year):
        """Return the formatted list that can be passed to a coordinate chart.

        Args:
            month_lists: list of daily values where each sublist is one month starting with January
            year: year expressed in 4 decimal places (i.e. 2019)

        Returns:
            list: of values with additional None values to align with grid

        """
        values = []
        for idx_month, daily_list in enumerate(month_lists):
            idx_first_day, count_days = calendar.monthrange(year, idx_month + 1)
            idx_first_day += 1  # Increment to start on Sunday -- PLANNED: make this configureable
            values.extend([None] * idx_first_day)
            values.extend(daily_list)
            values.extend([None] * (len(self.corners['x']) - idx_first_day - count_days))
        return values


class MonthGrid(GridClass):  # noqa: H601
    """Coordinates of days within a single month."""

    marker_kwargs = {'size': 35, 'symbol': 'square'}
    """Marker keyword arguments. Default is `{'size': 35, 'symbol': 'square'}`"""

    def __init__(self, grid_dims=(1, 1), titles=None):
        """Initialize the coordinates.

        Args:
            grid_dims: tuple of the number of tiles in grid. In format `(row, column)`. Default is (1, 1)
            titles: list of titles to place in each grid element. Default is None

        Raises:
            RuntimeError: if error in the grid dimensions or titles

        """
        if grid_dims != (1, 1):  # pragma: no cover
            raise RuntimeError('Day grid can only show one month, expected (1, 1)')
        if titles is not None and len(titles) != 1:  # pragma: no cover
            raise RuntimeError(f'Only one title is allowed for the MonthGrid. Received: {titles}')

        super().__init__(grid_dims=grid_dims, titles=titles)

        # Calculate four corners
        self.corners = calculate_calendar_grid_corners(margin=1.25)

    def format_data(self, daily_values, year, month):
        """Return the formatted list that can be passed to a coordinate chart.

        Args:
            daily_values: list of values for each day of month
            year: year expressed in 4 digits (2019, 2020, etc.)
            month: month index in [1, 12]

        Returns:
            list: of values with additional None values to align with grid

        """
        idx_first_day = calendar.monthrange(year, month)[0]
        values = [None] * idx_first_day
        values.extend(daily_values)
        return values



"""dash_charts."""

from loguru import logger

__version__ = '0.1.0'
__pkg_name__ = 'dash_charts'

logger.disable(__pkg_name__)

# ====== Above is the recommended code from calcipy_template and may be updated on new releases ======



"""Equations used in scipy fit calculations."""

import numpy as np


def linear(x_values, factor_a, factor_b):
    """Return result(s) of linear equation with factors of a and b.

    `y = a * x + b`

    Args:
        x_values: single number of list of numbers
        factor_a: number, slope
        factor_b: number, intercept

    Returns:
        y_values: as list or single digit

    """
    return np.add(
        np.multiply(factor_a, x_values),
        factor_b,
    )


def quadratic(x_values, factor_a, factor_b, factor_c):
    """Return result(s) of quadratic equation with factors of a, b, and c.

    `y = a * x^2 + b * x + c`

    Args:
        x_values: single number of list of numbers
        factor_a: number
        factor_b: number
        factor_c: number

    Returns:
        y_values: as list or single digit

    """
    return np.add(
        np.multiply(
            factor_a,
            np.power(x_values, 2),
        ),
        np.add(
            np.multiply(factor_b, x_values),
            factor_c,
        ),
    )


def power(x_values, factor_a, factor_b):
    """Return result(s) of quadratic equation with factors of a and b.

    `y = a * x^b`

    Args:
        x_values: single number of list of numbers
        factor_a: number
        factor_b: number

    Returns:
        y_values: as list or single digit

    """
    return np.multiply(
        factor_a,
        np.power(
            np.array(x_values).astype(float),
            factor_b,
        ),
    )


def exponential(x_values, factor_a, factor_b):
    """Return result(s) of exponential equation with factors of a and b.

    `y = a * e^(b * x)`

    Args:
        x_values: single number of list of numbers
        factor_a: number
        factor_b: number

    Returns:
        y_values: as list or single digit

    """
    return np.multiply(
        factor_a,
        np.exp(
            np.multiply(factor_b, x_values),
        ),
    )


def double_exponential(x_values, factor_a, factor_b, factor_c, factor_d):
    """Return result(s) of a double exponential equation with factors of a, b, c, and d.

    `y = a * e^(b * x) - c * e^(d * x)`

    Args:
        x_values: single number of list of numbers
        factor_a: number
        factor_b: number
        factor_c: number
        factor_d: number

    Returns:
        y_values: as list or single digit

    """
    return np.subtract(
        np.multiply(
            factor_a,
            np.exp(
                np.multiply(factor_b, x_values),
            ),
        ),
        np.multiply(
            factor_c,
            np.exp(
                np.multiply(factor_d, x_values),
            ),
        ),
    )



"""Utilities for better Dash callbacks."""

import re

import dash
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate


def format_app_callback(lookup, outputs, inputs, states):
    """Format list of [Output, Input, State] for `@app.callback()`.

    Args:
        lookup: dict with app_id keys that map to a globally unique component id
        outputs: list of tuples with app_id and property name
        inputs: list of tuples with app_id and property name
        states: list of tuples with app_id and property name

    Returns:
        list: list[lists] in order `(Outputs, Inputs, States)` for `@app.callback()`. Some sublists may be empty

    """
    return (
        [Output(component_id=lookup[_id], component_property=prop) for _id, prop in outputs],
        [Input(component_id=lookup[_id], component_property=prop) for _id, prop in inputs],
        [State(component_id=lookup[_id], component_property=prop) for _id, prop in states],
    )


def map_args(raw_args, inputs, states):
    """Map the function arguments into a dictionary with keys for the input and state names.

    For situations where the order of inputs and states may change, use this function to verbosely define the inputs:

    ```python
    a_in, a_state = map_args(raw_args, inputs, states)
    click_data = a_in[self.id_main_figure]['clickData']
    n_clicks = a_in[self.id_randomize_button]['n_clicks']
    data_cache = a_state[self.id_store]['data']
    ```

    Alternatively, for use cases that are unlikely to change the order of Inputs/State, unwrap positionally with:

    ```python
    click_data, n_clicks = args[:len(inputs)]
    data_cache = args[len(inputs):]
    ```

    Args:
        raw_args: list of arguments passed to callback
        inputs: list of input components. May be empty list
        states: list of state components. May be empty list

    Returns:
        dict: with keys of the app_id, property, and arg value (`a_in[key][arg_type]`)

    """
    # Split args into groups of inputs/states
    a_in = raw_args[:len(inputs)]
    a_state = raw_args[len(inputs):]

    # Map args into dictionaries
    arg_map = [{app_id: [] for app_id in {items[0] for items in group}} for group in [inputs, states]]
    for group_idx, (groups, args) in enumerate([(inputs, a_in), (states, a_state)]):
        # Assign the arg to the appropriate dictionary in arg_map
        for arg_idx, (app_id, prop) in enumerate(groups):
            arg_map[group_idx][app_id].append((prop, args[arg_idx]))
        for app_id in arg_map[group_idx].keys():
            arg_map[group_idx][app_id] = dict(arg_map[group_idx][app_id])
    return arg_map


def map_outputs(outputs, element_info):
    """Return properly ordered list of new Dash elements based on the order of outputs.

    Alternatively, for simple cases of 1-2 outputs, just return the list with:

    ```python
    return [new_element_1, new_element_2]
    ```

    Args:
        outputs: list of output components
        element_info: list of tuples with keys `(app_id, prop, element)`

    Returns:
        list: ordered list to match the order of outputs

    Raises:
        RuntimeError: Check that the number of outputs and the number of element_info match

    """
    if len(outputs) != len(element_info):
        raise RuntimeError(f'Expected same number of items between:\noutputs:{outputs}\nelement_info:{element_info}')

    # Create a dictionary of the elements
    lookup = {app_id: [] for app_id in {info[0] for info in element_info}}
    for app_id, prop, element in element_info:
        lookup[app_id].append((prop, element))
    for app_id in lookup:
        lookup[app_id] = dict(lookup[app_id])

    return [lookup[app_id][prop] for app_id, prop in outputs]


def get_triggered_id():
    """Use Dash context to get the id of the input element that triggered the callback.

    See advanced callbacks: https://dash.plotly.com/advanced-callbacks

    Returns:
        str: id of the input that triggered the callback

    Raises:
        PreventUpdate: if callback was fired without an input

    """
    ctx = dash.callback_context
    if not ctx.triggered:
        raise PreventUpdate

    prop_id = ctx.triggered[0]['prop_id']  # in format: `id.key` where we only want the `id`
    return re.search(r'(^.+)\.[^\.]+$', prop_id).group(1)



"""Helpers for building Dash applications."""

from contextlib import ContextDecorator

import dataset
import pandas as pd

from .utils_data import SQLConnection, uniq_table_id, write_csv

# ----------------------------------------------------------------------------------------------------------------------
# dataset

META_TABLE_NAME = 'meta'
"""Name of the Meta-Data table in a typical SQLite database."""


class DBConnect:  # noqa: H601
    """Manage database connection since closing connection isn't possible."""

    db_path = None
    """Path to the local storage SQLite database file. Initialize in `__init__()`."""

    _db = None

    @property
    def db(self):
        """Return connection to database. Will create new connection if one does not exist already.

        Returns:
            dict: `dataset` database instance

        """
        if self._db is None:
            self._db = dataset.connect(f'sqlite:///{self.db_path}')
        return self._db

    def __init__(self, db_path):
        """Store the database path and ensure the parent directory exists.

        Args:
            db_path: Path to the SQLite file

        """
        self.db_path = db_path.resolve()
        self.db_path.parent.mkdir(exist_ok=True)
        self.db  # Check initial connection

    def new_table(self, table_name):
        """Create a table. Drop a table if one existed before.

        Args:
            table_name: string table name to create

        Returns:
            table: a dataset Table instance

        """
        if table_name in self.db.tables:
            self.db[table_name].drop()
        return self.db.create_table(table_name)

    def close(self):
        """Safely disconnect and release the SQLite file."""
        self.db.executable.close()
        self._db = None


class DBConnection(ContextDecorator):
    """Ensure the DBConnect connection is properly opened and closed."""

    def __init__(self, db_path):
        """Initialize context wrapper.

        Args:
            db_path: Path to the SQLite file

        """
        self.conn = None
        self.db_path = db_path

    def __enter__(self):
        """Connect to the database and return connection reference.

        Returns:
            dict: connection to sqlite database

        """
        self.conn = DBConnect(self.db_path)
        return self.conn

    def __exit__(self, exc_type, exc_value, traceback):
        """Close connection."""  # noqa: DAR101
        self.conn.close()


def export_table_as_csv(csv_filename, table):
    """Create a CSV file summarizing a table of a `dataset` database.

    Args:
        csv_filename: Path to csv file
        table: table from dataset database

    """
    rows = [[*table.columns]]
    rows.extend([*row.values()] for row in table)
    write_csv(csv_filename, rows)


def safe_col_name(args_pair):
    """Ensure that the column name is safe for SQL (unique value, no spaces, no trailing punctuation).

    Typically called with `df.columns = [*map(safe_col_name, enumerate(df.columns.to_list()))]`

    Args:
        args_pair: tuple of arguments from map function in `(idx, col)`

    Returns:
        string: safely formatted string for SQLite

    """
    idx, col = args_pair
    col = col.strip().replace(' ', '_').replace('.', '_').replace(',', '_')
    return str(idx) if col == '' else col


def store_reference_tables(db_path, data_dicts, meta_table_name=META_TABLE_NAME, use_raw_sql=True):   # noqa: CCR001
    """Store multi-dimensionsal data in a SQLite database.

    WARN: This will append to the META_TABLE_NAME without checking for duplicates. Handling de-duping separately

    Args:
        db_path: Path to a `.db` file
        data_dicts: all data to be stored in SQLite. Can contain Pandas dataframes
        meta_table_name: optional name of the main SQLite table. Default is `META_TABLE_NAME`
        use_raw_sql: if True, will use the raw SQL connection rather than DataSet. This is faster for meta_tables that
            have more than 1000 rows, but less safe

    """
    with SQLConnection(db_path) as conn:
        meta_table = []
        unique = uniq_table_id()
        for dict_idx, data_dict in enumerate(data_dicts):
            lookup = {}
            for key_idx, (key, value) in enumerate(data_dict.items()):
                if isinstance(value, pd.DataFrame):
                    value.columns = [*map(safe_col_name, enumerate(value.columns.to_list()))]
                    table_name = f'{unique}Dict{dict_idx}Key{key_idx}'
                    value.to_sql(table_name, con=conn)
                    lookup[key] = table_name
                else:
                    lookup[key] = value
            meta_table.append(lookup)

    if use_raw_sql:
        add_meta_table_records_sql(db_path, meta_table, meta_table_name)
    else:
        with DBConnection(db_path) as data_db:
            table_main = data_db.db.create_table(meta_table_name)
            table_main.insert_many(meta_table)


def add_meta_table_records_sql(db_path, meta_table, meta_table_name):
    """Store new rows for the meta table using a more performant SQLite implementation.

    WARN: This will append to the META_TABLE_NAME without checking for duplicates. Handling de-duping separately

    Args:
        db_path: Path to a `.db` file
        meta_table: list of dictionaries to add to the meta_table
        meta_table_name: optional name of the main SQLite table

    """
    with SQLConnection(db_path) as conn:
        cursor = conn.cursor()
        keys = [*meta_table[0].keys()]
        names_formatted = ','.join(map(safe_col_name, enumerate(keys)))
        cursor.execute(f'CREATE TABLE IF NOT EXISTS {meta_table_name}({names_formatted});')
        rows = [[row[col] for col in keys] for row in meta_table]
        places = ','.join(['?'] * len(keys))
        cursor.executemany(f'INSERT INTO {meta_table_name}({names_formatted}) VALUES ({places});', rows)
        conn.commit()


def get_table(db_path, table_name, drop_id_col=True):
    """Retrieve the meta table as a Pandas dataframe.

    Args:
        db_path: Path to a `.db` file
        table_name: SQLite table name
        drop_id_col: if True, drop the `id` column from SQL. Default is True

    Returns:
        df_table: pandas dataframe for the values in the specified table (`meta_table_name`)

    """
    with DBConnection(db_path) as data_db:
        df_table = pd.DataFrame([*data_db.db[table_name].all()])
    # Optionally remove the 'id' column added in the SQL database
    if drop_id_col:
        df_table = df_table.drop(labels=['id', 'index'], axis=1, errors='ignore')
    return df_table  # noqa: R504



"""Gantt Chart.

Note: does not support resources nor task dependencies; however those could be added by extending this base class.

# Removed Code

```py
# Just snippets of Python code that may be useful in the future
dates = sorted(set(filter(None, df_raw['start'].to_list() + df_raw['end'].to_list())))
self.axis_range = {'x': [dates[0], dates[-1]]}
```

"""

import plotly.graph_objects as go
from palettable.tableau import TableauMedium_10

from .utils_data import format_unix, get_unix
from .utils_fig import CustomChart


class GanttChart(CustomChart):  # noqa: H601
    """Gantt Chart: task and milestone timeline."""

    date_format = '%Y-%m-%d'
    """Date format for bar chart."""

    pallette = TableauMedium_10.hex_colors
    """Default color pallette for project colors."""

    hover_label_settings = {'bgcolor': 'white', 'font_size': 12, 'namelength': 0}
    """Plotly hover label settings."""

    rh = 1
    """Height of each rectangular task."""

    def create_traces(self, df_raw):
        """Return traces for plotly chart.

        Args:
            df_raw: pandas dataframe with columns: `(category, label, start, end, progress)`

        Returns:
            list: Dash chart traces

        """
        # If start is None, assign end to start so that the sort is correct
        start_index = df_raw.columns.get_loc('start')
        end_index = df_raw.columns.get_loc('end')
        for index in [idx for idx, is_na in enumerate(df_raw['start'].isna()) if is_na]:
            df_raw.iloc[index, start_index] = df_raw.iloc[index, end_index]
        df_raw['progress'] = df_raw['progress'].fillna(0)  # Fill possibly missing progress values for milestones
        df_raw = (
            df_raw
            .sort_values(by=['category', 'start'], ascending=False)
            .sort_values(by=['end'], ascending=False)
            .reset_index(drop=True)
        )
        # Create color lookup using categories in sorted order
        categories = set(df_raw['category'])
        self.color_lookup = {cat: self.pallette[idx] for idx, cat in enumerate(categories)}
        # Track which categories have been plotted
        plotted_categories = []
        # Create the Gantt traces
        traces = []
        for task in df_raw.itertuples():
            y_pos = task.Index
            is_first = task.category not in plotted_categories
            plotted_categories.append(task.category)
            traces.append(self._create_task_shape(task, y_pos, is_first))
            if task.progress > 0:
                traces.append(self._create_progress_shape(task, y_pos))
            traces.append(self._create_annotation(task, y_pos))
        return traces

    def _create_hover_text(self, task):
        """Return hover text for given trace.

        Args:
            task: row tuple from df_raw with: `(category, label, start, end, progress)`

        Returns:
            string: HTML-formatted hover text

        """
        dates = [format_unix(get_unix(str_ts, self.date_format), '%a, %d%b%Y') for str_ts in [task.start, task.end]]
        if task.start != task.end:
            date_range = f'<br><b>Start</b>: {dates[0]}<br><b>End</b>: {dates[1]}'
        else:
            date_range = f'<br><b>Milestone</b>: {dates[1]}'
        return f'<b>{task.category}</b><br>{task.label} ({int(task.progress * 100)}%)<br>{date_range}'

    def _create_task_shape(self, task, y_pos, is_first):
        """Create colored task scatter rectangle.

        Args:
            task: row tuple from df_raw with: `(category, label, start, end, progress)`
            y_pos: top y-coordinate of task
            is_first: if True, this is the first time a task of this category will be plotted

        Returns:
            trace: single Dash chart Scatter trace

        """
        color = self.color_lookup[task.category]
        scatter_kwargs = {
            'fill': 'toself',
            'fillcolor': color,
            'hoverlabel': self.hover_label_settings,
            'legendgroup': color,
            'line': {'width': 1},
            'marker': {'color': color},
            'mode': 'lines',
            'showlegend': is_first,
            'text': self._create_hover_text(task),
            'x': [task.start, task.end, task.end, task.start, task.start],
            'y': [y_pos, y_pos, y_pos - self.rh, y_pos - self.rh, y_pos],
        }
        if is_first:
            scatter_kwargs['name'] = task.category
        return go.Scatter(**scatter_kwargs)

    def _create_progress_shape(self, task, y_pos):
        """Create semi-transparent white overlay `self.shapes` to indicate task progress.

        Args:
            task: row tuple from df_raw with: `(category, label, start, end, progress)`
            y_pos: top y-coordinate of task

        Returns:
            trace: single Dash chart Scatter trace

        """
        unix_start = get_unix(task.start, self.date_format)
        unix_progress = (get_unix(task.end, self.date_format) - unix_start) * task.progress + unix_start
        end = format_unix(unix_progress, self.date_format)
        return go.Scatter(
            fill='toself',
            fillcolor='white',
            hoverinfo='skip',
            legendgroup=self.color_lookup[task.category],
            line={'width': 1},
            marker={'color': 'white'},
            mode='lines',
            opacity=0.5,
            showlegend=False,
            x=[task.start, end, end, task.start, task.start],
            y=[y_pos, y_pos, y_pos - self.rh, y_pos - self.rh, y_pos],
        )

    def _create_annotation(self, task, y_pos):
        """Add task label to chart as text overlay.

        Args:
            task: row tuple from df_raw with: `(category, label, start, end, progress)`
            y_pos: top y-coordinate of task

        Returns:
            trace: single Dash chart Scatter trace

        """
        # For milestones with narrow fill, hover can be tricky, so intended to make the whole length of the text
        #   hoverable, but only the x/y point appears to be hoverable although it makes a larger hover zone at least
        return go.Scatter(
            hoverlabel=self.hover_label_settings,
            hovertemplate=self._create_hover_text(task) + '<extra></extra>',
            hovertext=self._create_hover_text(task),
            legendgroup=self.color_lookup[task.category],
            mode='text',
            showlegend=False,
            text=task.label,
            textposition='middle left',
            x=[task.end],
            y=[y_pos - self.rh / 2],
        )

    def create_layout(self):
        """Extend the standard layout.

        Returns:
            dict: layout for Dash figure

        """
        layout = super().create_layout()
        # Suppress Y axis ticks/grid
        layout['yaxis']['showgrid'] = False
        layout['yaxis']['showticklabels'] = False
        layout['yaxis']['zeroline'] = False
        return layout



"""Classes for more complex applications that have tabbed or paged navigation."""

from collections import OrderedDict
from copy import deepcopy

import dash_bootstrap_components as dbc
from dash import dcc, html
from implements import implements

from .utils_app import AppBase, AppInterface

TODO_CLIENT_CALLBACK = '''
TODO: Create clientside callbacks dynamically to update the title on navigation
See: http://dash.plotly.com/external-resources
```py
app.clientside_callback(
    """
    function(tab_value) {
        if (tab_value === 'tab-1') {
            document.title = 'Tab 1'
        } else if (tab_value === 'tab-2') {
            document.title = 'Tab 2'
        }
    }
    """,
    Output('blank-output', 'children'),
    [Input('tabs-example', 'value')]
)
```
'''


# TODO: Try to see if I can resolve the interface differences or if I need make a subclass interface
# @implements(AppInterface)  # noqa: H601
class AppWithNavigation(AppBase):
    """Base class for building Dash Application with tabs or URL routing."""

    app = None
    """Main Dash application to pass to all child tabs."""

    nav_lookup = None
    """OrderedDict based on the list of tuples from `self.define_nav_elements()`."""

    nav_layouts = None
    """Dictionary with nav_names as keys and corresponding layout as value."""

    def define_nav_elements(self):
        """Return list of initialized pages or tabs accordingly.

        Should return, list: each item is an initialized app (ex `[AppBase(self.app)]` in the order each tab is rendered

        Raises:
            NotImplementedError: Child class must implement this method

        """
        raise NotImplementedError('define_nav_elements must be implemented by child class')  # pragma: no cover

    def create(self, **kwargs):
        """Create each navigation componet, storing the layout. Then parent class to create application.

        Args:
            kwargs: keyword arguments passed to `self.create`

        """
        # Initialize the lookup for each tab then configure each tab
        self.nav_lookup = OrderedDict([(tab.name, tab) for tab in self.define_nav_elements()])
        self.nav_layouts = {}
        for nav_name, nav in self.nav_lookup.items():
            nav.create(assign_layout=False)
            self.nav_layouts[nav_name] = nav.return_layout()

        # Store validation_layout that is later used for callback verification in base class
        self.validation_layout = [*map(deepcopy, self.nav_layouts.values())]

        # Initialize parent application that handles navigation
        super().create(**kwargs)

    def initialization(self) -> None:
        """Initialize ids with `self.register_uniq_ids([...])` and other one-time actions."""
        super().initialization()
        self.register_uniq_ids(self.app_ids)

    def create_elements(self) -> None:
        """Override method as not needed at navigation-level."""
        ...  # pragma: no cover

    def create_callbacks(self) -> None:
        """Override method as not needed at navigation-level."""
        ...  # pragma: no cover


@implements(AppInterface)  # noqa: H601
class StaticTab(AppBase):
    """Simple App without charts or callbacks."""

    basic_style = {
        'marginLeft': 'auto',
        'marginRight': 'auto',
        'maxWidth': '1000px',
        'paddingTop': '10px',
    }

    def initialization(self) -> None:
        """Initialize ids with `self.register_uniq_ids([...])` and other one-time actions."""
        super().initialization()
        self.register_uniq_ids(['N/A'])

    def create_elements(self) -> None:
        """Initialize the charts, tables, and other Dash elements.."""
        ...

    def create_callbacks(self) -> None:
        """Register callbacks necessary for this tab."""
        ...


class AppWithTabs(AppWithNavigation):
    """Base class for building Dash Application with tabs."""

    # App ids
    id_tabs_content = 'tabs-wrapper'
    id_tabs_select = 'tabs-content'

    app_ids = [id_tabs_content, id_tabs_select]
    """List of all ids for the top-level tab view. Will be mapped to `self._il` for globally unique ids."""

    def return_layout(self) -> dict:
        """Return Dash application layout.

        Returns:
            dict: Dash HTML object

        """
        tabs = [dcc.Tab(label=name, value=name) for name, tab in self.nav_lookup.items()]
        return html.Div(
            children=[
                dcc.Tabs(
                    id=self._il[self.id_tabs_select], value=list(self.nav_lookup.keys())[0],
                    children=tabs,
                ),
                html.Div(id=self._il[self.id_tabs_content]),
            ],
        )

    def create_callbacks(self) -> None:
        """Register the navigation callback."""
        outputs = [(self.id_tabs_content, 'children')]
        inputs = [(self.id_tabs_select, 'value')]

        @self.callback(outputs, inputs, [])
        def render_tab(tab_name):
            return [self.nav_layouts[tab_name]]


# > PLANNED: Make the tabs and chart compact as well when the compact argument is set to True
class FullScreenAppWithTabs(AppWithTabs):  # noqa: H601
    """Base class for building Dash Application with tabs that uses the full window."""

    tabs_location = 'left'
    """Tab orientation setting. One of `(left, top, bottom, right)`."""

    tabs_margin = '10%'
    """Adjust this setting based on the width or height of the tabs to prevent the content from overlapping the tabs."""

    tabs_compact = False
    """Boolean setting to toggle between a padded tab layout if False and a minimal compact version if True."""

    def verify_app_initialization(self):
        """Check that the app was properly initialized.

        Raises:
            RuntimeError: if child class has not called `self.register_uniq_ids`

        """
        super().verify_app_initialization()
        allowed_locations = ('left', 'top', 'bottom', 'right')
        if self.tabs_location not in allowed_locations:  # pragma: no cover
            raise RuntimeError(f'`self.tabs_location = {self.tabs_location}` is not in {allowed_locations}')

    def return_layout(self) -> dict:
        """Return Dash application layout.

        Returns:
            dict: Dash HTML object

        """
        return html.Div(
            children=[
                self.tab_menu(),
                html.Div(
                    style={f'margin-{self.tabs_location}': self.tabs_margin},
                    children=[html.Div(id=self._il[self.id_tabs_content])],
                ),
            ],
        )

    def generate_tab_kwargs(self):
        """Create the tab keyword arguments. Intended to be modified through inheritance.

        Returns:
            tuple: keyword arguments and styling for the dcc.Tab elements

                - tab_kwargs: with at minimum keys `(style, selected_style)` for dcc.Tab
                - tabs_kwargs: to be passed to dcc.Tabs
                - tabs_style: style for the dcc.Tabs HTML element

        """
        # Unselected tab style
        if self.tabs_compact:
            tab_style = {'padding': '2px 4px 2px 4px'}
            tabs_padding = '6px 0 0 2px'
        else:
            tab_style = {'padding': '10px 20px 10px 20px'}
            tabs_padding = '15px 0 0 5px'
        # Extend tab style for selected case
        selected_style = deepcopy(tab_style)
        opposite_lookup = {'top': 'bottom', 'bottom': 'top', 'left': 'right', 'right': 'left'}
        tabs_style = {   # noqa: ECE001
            'backgroundColor': '#F9F9F9',
            'padding': tabs_padding,
            'position': 'fixed',
            'zIndex': '999',
            f'border{opposite_lookup[self.tabs_location].title()}': '1px solid #d6d6d6',
            self.tabs_location: '0',
        }
        if self.tabs_location in ['left', 'right']:
            # Configure for vertical case
            selected_style['border-left'] = '3px solid #119DFF'
            tabs_kwargs = {
                'vertical': True,
                'style': {'width': '100%'},
                'parent_style': {'width': '100%'},
            }
            tabs_style['top'] = '0'
            tabs_style['bottom'] = '0'
            tabs_style['width'] = 'auto'
        else:
            # Configure for horizontal case
            selected_style['border-top'] = '3px solid #119DFF'
            tabs_kwargs = {}
            tabs_style['height'] = 'auto'
            tabs_style['right'] = '0'
            tabs_style['left'] = '0'

        tab_kwargs = {'style': tab_style, 'selected_style': selected_style}
        return (tab_kwargs, tabs_kwargs, tabs_style)

    def tab_menu(self):
        """Return the HTML elements for the tab menu.

        Returns:
            dict: Dash HTML object

        """
        tab_kwargs, tabs_kwargs, tabs_style = self.generate_tab_kwargs()
        tabs = [dcc.Tab(label=name, value=name, **tab_kwargs) for name, tab in self.nav_lookup.items()]
        return html.Div(
            children=[
                dcc.Tabs(
                    id=self._il[self.id_tabs_select], value=list(self.nav_lookup.keys())[0],
                    children=tabs, **tabs_kwargs,
                ),
            ], style=tabs_style,
        )


class AppMultiPage(AppWithNavigation):  # noqa: H601
    """Base class for building Dash Application with multiple pages."""

    navbar_links = None
    """Base class must create list of tuples `[('Link Name', '/link'), ]` to use default `self.nav_bar()`."""

    dropdown_links = None
    """Base class must create list of tuples `[('Link Name', '/link'), ]` to use default `self.nav_bar()`."""

    logo = None
    """Optional path to logo. If None, no logo will be shown in navbar."""

    # App ids
    id_url = 'pages-url'
    id_pages_content = 'pages-wrapper'
    id_toggler = 'nav-toggle'
    id_collapse = 'nav-collapse'

    app_ids = [id_url, id_pages_content, id_toggler, id_collapse]
    """List of all ids for the top-level pages view. Will be mapped to `self._il` for globally unique ids."""

    def return_layout(self) -> dict:
        """Return Dash application layout.

        Returns:
            dict: Dash HTML object

        """
        return html.Div(
            children=[
                dcc.Location(id=self._il[self.id_url], refresh=False),
                self.nav_bar(),
                html.Div(id=self._il[self.id_pages_content]),
            ],
        )

    def nav_bar(self):
        """Return the HTML elements for the navigation menu.

        Returns:
            dict: Dash HTML object

        """
        # Create brand icon and name where icon in optional
        brand = []
        if self.logo:
            brand.append(dbc.Col(html.Img(src=self.logo, height='25px')))
        brand.append(dbc.Col(dbc.NavbarBrand(self.name, className='ml-2')))
        # Create links in navbar and dropdown. Both are optional
        links = []
        if self.navbar_links:
            links.append(
                dbc.Nav(
                    children=[dbc.NavItem(dbc.NavLink(name, href=link)) for name, link in self.navbar_links],
                    fill=True,
                    navbar=True,
                ),
            )
        if self.dropdown_links:
            links.append(
                dbc.Nav(
                    dbc.DropdownMenu(
                        children=[dbc.DropdownMenuItem(name, href=link) for name, link in self.dropdown_links],
                        in_navbar=True,
                        label='Links',
                        nav=True,
                    ),
                    navbar=True,
                ),
            )
        # Layout default navbar
        return dbc.Navbar(
            children=[
                dbc.NavLink(
                    [
                        dbc.Row(
                            children=brand,
                            align='center g-0',
                        ),
                    ], href='/',
                ),
                dbc.NavbarToggler(id=self._il[self.id_toggler]),
                dbc.Collapse(
                    dbc.Row(
                        children=links,
                        className='flex-nowrap mt-3 mt-md-0 g-0',
                        align='center',
                    ),
                    id=self._il[self.id_collapse],
                    navbar=True,
                ),
            ],
            sticky='top',
            color='dark',
            dark=True,
        )

    def create_callbacks(self) -> None:
        """Register the navigation callback."""
        outputs = [(self.id_pages_content, 'children')]
        inputs = [(self.id_url, 'pathname')]

        @self.callback(outputs, inputs, [])
        def render_page(pathname):
            try:
                # TODO: Demo how pages could use parameters from pathname
                return [self.nav_layouts[self.select_page_name(pathname)]]
            except Exception as err:
                return [html.Div(children=[f'Error rendering "{pathname}":\n{err}'])]

        @self.callback(
            [(self.id_collapse, 'is_open')],
            [(self.id_toggler, 'n_clicks')],
            [(self.id_collapse, 'is_open')],
        )
        def toggle_navbar_collapse(n_clicks, is_open):
            return [not is_open if n_clicks else is_open]

    def select_page_name(self, pathname):
        """Return the page name determined based on the pathname.

        Should return str: page name

        Args:
            pathname: relative pathname from URL

        Raises:
            NotImplementedError: Child class must implement this method

        """
        raise NotImplementedError('nav_bar must be implemented by child class')  # pragma: no cover



"""DataTable Base Classes."""

from dash import dash_table

# TODO: See pattern mathing callbacks for adding buttons (to show modal) to datatables
#   https://dash.plotly.com/pattern-matching-callbacks

# PLANNED: see conditional formatting: https://dash.plotly.com/datatable/conditional-formatting

# PLANNED: These methods may be replaced in a future version of Dash
# Currently, edge case when column is string, but filter could be a number
# See: https://dash.plot.ly/datatable/callbacks & https://github.com/plotly/dash-table/issues/441

OPERATORS = [
    ['ge ', '>='],
    ['le ', '<='],
    ['lt ', '<'],
    ['gt ', '>'],
    ['ne ', '!='],
    ['eq ', '='],
    ['contains '],
    ['datestartswith '],
]
"""List of lists containing each possible filter string."""


def split_filter_part(filter_part):  # noqa: CCR001
    """Split the filter into `(name, operator, value)` components.

    Based on `Backend Paging with Filtering`: https://dash.plot.ly/datatable/callbacks

    Args:
        filter_part: string filter query

    Returns:
        tuple: `(name, operator, value)` which could be all None if no match was found

    """
    for operator_type in OPERATORS:
        for operator in operator_type:
            if operator in filter_part:
                name_part, value_part = filter_part.split(operator, 1)
                name = name_part[name_part.find('{') + 1: name_part.rfind('}')]

                value_part = value_part.strip()
                v0 = value_part[0]
                if (v0 == value_part[-1] and v0 in ("'", '"', '`')):
                    value = value_part[1: -1].replace(f'\\{v0}', v0)
                else:
                    try:
                        value = float(value_part)
                    except ValueError:
                        value = value_part

                # Word operators need spaces after them in the filter string, but we don't want these later
                return (name, operator_type[0].strip(), value)

    return [None, None, None]


def apply_datatable_filters(df_table, filter_query):
    """Filter a dataframe based on Dash datatable filterquery.

    Based on `Backend Paging with Filtering`: https://dash.plot.ly/datatable/callbacks

    Args:
        df_table: pandas dataframe to filter
        filter_query: Dash datatable string filter query

    Returns:
        dataframe: filtered dataframe

    """
    filtering_expressions = filter_query.split(' && ')
    for filter_part in filtering_expressions:
        col_name, operator, filter_value = split_filter_part(filter_part)
        if operator in ('eq', 'ne', 'lt', 'le', 'gt', 'ge'):
            # these operators match pandas series operator method names
            df_table = df_table.loc[getattr(df_table[col_name], operator)(filter_value)]
        elif operator == 'contains':
            df_table = df_table.loc[df_table[col_name].str.contains(filter_value)]
        elif operator == 'datestartswith':
            # this is a simplification of the front-end filtering logic,
            # only works with complete fields in standard format
            df_table = df_table.loc[df_table[col_name].str.startswith(filter_value)]
    return df_table


# PLANNED: Maybe move parameters to attr.ib classes?
class BaseDataTable:  # noqa: H601
    """Base Class for Data Tables."""

    filter_summary = """Table Filter Syntax:

Full documentation at: [https://dash.plot.ly/datatable/filtering](https://dash.plot.ly/datatable/filtering)

- `eq`: exact match (applies to number columns and will try to convert string to number)
- `contains`: search for exact (case-sensitive) substring in each cell
- `datestartswith`: matches partial datetime. For example, datestartswith `2018-03-01` will match `2018-03-01 12:59`
but not `2018-03`
- `ne`, `gt`, `ge`, `lt`, `le`: comparison operators for not equal, greater than, greater or equal, less than, etc.
Applies to numbers and string columns (uses numbers, symbols, uppercase letter, lowercase letters)

Press enter of tab to apply the filter"""
    """Markdown text explaining dash_table.DataTable filtering rules with link to full documentation."""

    # dash_table.DataTable Parameters. Documentation: https://dash.plot.ly/datatable/reference

    style_table = None
    """DataTable.style_table dictionary. Default enables overflowX scroll. Set in `initialize_mutables`."""

    css = None
    """DataTable.css list. Use the style_* properties first.  Set in `initialize_mutables`.

    Default sets row margin to zero to fix a negative margin issue when using dash_table and Bootstrap
    See: https://github.com/facultyai/dash-bootstrap-components/issues/334

    Also sets other various style tweaks to highlight the filter icons on hover, etc.

    """

    style_cell = None
    """DataTable.style_cell dictionary. Default is empty dictionary. Set in `initialize_mutables`."""

    style_cell_conditional = None
    """DataTable.style_cell_conditional list. Default is empty list. Set in `initialize_mutables`."""

    style_data = None
    """DataTable.style_data dictionary. Default is empty dictionary. Set in `initialize_mutables`."""

    style_data_conditional = None
    """DataTable.style_data_conditional list. Default is for odd rows to have an off-white background (zebra stripe).

    Set in `initialize_mutables`

    """

    style_header = None
    """DataTable.style_header dictionary. Default is bold and off-white background. Set in `initialize_mutables`."""

    style_header_conditional = None
    """DataTable.style_header_conditional list. Default is empty list. Set in `initialize_mutables`."""

    style_filter = None
    """DataTable.style_filter dictionary. Default is empty dictionary. Set in `initialize_mutables`."""

    style_filter_conditional = None
    """DataTable.style_filter_conditional list. Default is empty list. Set in `initialize_mutables`."""

    column_selectable = None
    """DataTable.column_selectable. Default is `None`."""

    column_kwarg_lookup = None
    """Lookup for keyword arguments for each column allowing deletable, selectable, etc. to be set per column."""

    export_format = 'none'
    """DataTable.export_format. Default is `'none'`. Could be one of `(csv, xlsx)`."""

    export_headers = 'names'
    """DataTable.export_headers. Default is `'names'`. See documentation."""

    filter_action = 'native'
    """DataTable.filter_action. Default is `'native'`."""

    page_size = 25
    """DataTable.page_size. Default is `'25'`."""

    row_selectable = None
    """DataTable.row_selectable. Default is `None`."""

    style_as_list_view = False
    """DataTable.style_as_list_view. Default is False."""

    sort_action = 'native'
    """DataTable.sort_action. Default is `'native'`."""

    sort_mode = 'single'
    """DataTable.sort_mode. Default is `'single'`."""

    # CSS Variables

    text_color = '#333333'
    """Default text color."""

    background_color = '#ffffff'
    """Default background color."""

    zebra_color = '#f9f9f9'
    """Default background color for odd rows (zebra-stripe)."""

    selected_cell_color = '#eaeaea'
    """Default color for selected cells."""

    def __init__(self):
        """Initialize class."""
        self.initialize_mutables()

    def initialize_mutables(self):
        """Initialize the mutable data members to prevent modifying one attribute and impacting all instances."""
        self.style_table = {'overflowX': 'scroll'}
        self.css = [
            # Fix width when overflow is True and using dash_bootstrap_components
            {'selector': '.row', 'rule': 'margin: 0'},

            # Based on: https://community.plot.ly/t/dash-table-datatable-styling-examples/15594/
            # Align the select icon to the left and the sort icon to the right
            {
                'selector': 'th.dash-header div',
                'rule': 'display: flex; align-items: center;',
            },
            {
                'selector': 'th.dash-header div span.column-header--select',
                'rule': 'order: 1; flex-shrink: 5; text-align: left;',
            },
            {
                'selector': 'th.dash-header div span.column-header-name',
                'rule': 'order: 2; flex-grow: 5; text-align: center;',
            },
            {
                'selector': 'th.dash-header div span.column-header--sort',
                'rule': 'order: 3; flex-shrink: 5; text-align: right;',
            },

            # Remove excess borders to better match JQuery DataTables styling
            # (Some border styles don't appear to work from `self.style_header`)
            {'selector': 'th.dash-filter', 'rule': 'border-bottom-color: rgb(17, 17, 17) !important;'},
        ]
        self.style_cell = {
            'backgroundColor': self.background_color,
            'borderStyle': f'1px solid {self.zebra_color}',
            'color': self.text_color,
        }
        self.style_cell_conditional = []
        self.style_data = {}
        self.style_data_conditional = [
            {'if': {'row_index': 'even'}, 'backgroundColor': self.zebra_color},
        ]
        self.style_header = {
            'borderLeft': 'none',
            'borderRight': 'none',
            'borderTop': 'none',
            'fontWeight': 'bold',
        }
        self.style_header_conditional = []
        self.style_filter = {
            'borderBottomColor': 'rgb(17, 17, 17)',
        }
        self.style_filter_conditional = []

        self.initialize_column_kwarg_lookup()  # Must be called last

    def initialize_column_kwarg_lookup(self):
        """Initialize the column lookup based on the other mutables.

        Documentation: https://dash.plot.ly/datatable/reference

        Additional kwargs not set in default lookup:
            - clearable, deletable, editable, hideable, renamable, format, presentation, on_change, sort_as_null,
                validation, type

        """
        self.column_kwarg_lookup = {
            'default_kwargs': {
                'selectable': True,
            },
        }

    def create_table(self, df_raw, columns=None, **kwargs_datatable):
        """Create the dash_table.DataTable.

        Args:
            df_raw: data to pass to datatable
            columns: list of column names to display. Default is None to use all columns from df_raw
            kwargs_datatable: keyword arguments to pass to the datatable

        Returns:
            DataTable: returns dash_table.DataTable object

        """
        if columns is None:
            columns = df_raw.columns
        return dash_table.DataTable(**self._create_datatable(df_raw, columns), **kwargs_datatable)

    def format_datatable_columns(self, df_raw, columns):
        """Return a list of column names formatted for a dash_table. Uses `self.self.column_name_kwarg_lookup`.

        Args:
            df_raw: data to pass to datatable
            columns: list of strings or None

        Returns:
            list: of dict with keys `(name, id, deleteable, selectable)` in order of df_raw columns

        """
        return [
            {'name': col, 'id': col, **self.column_kwarg_lookup.get(col, self.column_kwarg_lookup['default_kwargs'])}
            for col in df_raw.columns if (columns is None or col in columns)
        ]

    def _create_datatable(self, df_raw, columns, **table_kwargs):
        """Return dictionary of keyword arguments for datatable.

        Args:
            df_raw: data to pass to datatable
            columns: will auto format list of string column names to display or use list of dicts
            table_kwargs: additional keyword arguments to pass to datatable, such as id

        Returns:
            dict: keys include `(columns, data)` and all data members

        """
        return {
            'columns': self.format_datatable_columns(df_raw, columns) if isinstance(columns[0], str) else columns,
            'data': (df_raw.loc[:, columns] if columns is not None else df_raw).to_dict('records'),

            # Add all datamembers
            'css': self.css,
            'style_table': self.style_table,
            'style_cell': self.style_cell,
            'style_cell_conditional': self.style_cell_conditional,
            'style_data': self.style_data,
            'style_data_conditional': self.style_data_conditional,
            'style_header': self.style_header,
            'style_header_conditional': self.style_header_conditional,
            'style_filter': self.style_filter,
            'style_filter_conditional': self.style_filter_conditional,
            'column_selectable': self.column_selectable,
            'export_format': self.export_format,
            'export_headers': self.export_headers,
            'filter_action': self.filter_action,
            'page_size': self.page_size,
            'row_selectable': self.row_selectable,
            'style_as_list_view': self.style_as_list_view,
            'sort_action': self.sort_action,
            'sort_mode': self.sort_mode,
        }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dash_charts']

package_data = \
{'': ['*'], 'dash_charts': ['assets/*']}

install_requires = \
['Palettable>=3.3.0',
 'astor>=0.8.1',
 'attrs-strict>=0.2.2',
 'beautifulsoup4>=4.10.0',
 'calcipy>=0.11.0',
 'cerberus>=1.3.4',
 'dash-bootstrap-components>=1.0.0',
 'dash-extensions>=0.0.55',
 'dash-tabulator>=0.4.2',
 'dash[testing]>=2.0.0',
 'dataset>=1.5.2',
 'dominate>=2.6.0',
 'implements>=0.3.0',
 'jsonpickle>=2.1.0',
 'lxml>=4.7.1',
 'markdown>=3.3.6',
 'numpy>=1.22.2',
 'pandas>=1.3.0',
 'psutil>=5.9.0',
 'python-box>=5.4.1',
 'scipy>=1.6.1',
 'tqdm>=4.62.3']

setup_kwargs = {
    'name': 'dash-charts',
    'version': '0.1.0',
    'description': 'Python package for Plotly/Dash apps with support for multi-page, modules, and new charts such as Pareto with an Object Orient Approach',
    'long_description': '# dash_charts\n\nPython package for [Plotly/Dash](https://dash.plot.ly/) apps with support for multi-page, modules, and new charts such as Pareto with an Object Orient Approach\n\nIncludes base classes for building a custom chart or application, new charts such as a Pareto, and base classes for tabbed or multi-page applications. See full documentation at [https://kyleking.me/dash_charts/](https://kyleking.me/dash_charts/). ( TODO: Currently not online )\n\n<!-- TOC -->\n\n- [dash_charts](#dash_charts)\n    - [Nov2020-Mar2021 Updates](#nov2020-mar2021-updates)\n    - [Quick Start](#quick-start)\n        - [1. Install](#1-install)\n        - [2. Example Code](#2-example-code)\n        - [3. Resulting Pareto Chart](#3-resulting-pareto-chart)\n        - [4. Additional Notes](#4-additional-notes)\n    - [Design Principles](#design-principles)\n    - [Local Development](#local-development)\n    - [Example Charts and Tables](#example-charts-and-tables)\n        - [Pareto Chart](#pareto-chart)\n        - [Gantt Chart](#gantt-chart)\n        - [Time Vis Chart](#time-vis-chart)\n        - [Rolling Mean and STD Chart](#rolling-mean-and-std-chart)\n        - [Fitted Chart](#fitted-chart)\n        - [Real Time SQL Demo](#real-time-sql-demo)\n        - [Coordinate Chart](#coordinate-chart)\n        - [Marginal Chart](#marginal-chart)\n        - [Data Table Module](#data-table-module)\n        - [Upload Module](#upload-module)\n    - [Applications](#applications)\n        - [Tabbed Application](#tabbed-application)\n        - [Multi-Page Application](#multi-page-application)\n        - [PX Generic Application](#px-generic-application)\n    - [Other](#other)\n        - [Cache](#cache)\n        - [Static HTML Generation](#static-html-generation)\n    - [Coverage](#coverage)\n    - [External Links](#external-links)\n    - [Installation](#installation)\n    - [Usage](#usage)\n    - [Roadmap](#roadmap)\n    - [Contributing](#contributing)\n    - [License](#license)\n\n<!-- /TOC -->\n\n## Nov2020-Mar2021 Updates\n\n<!-- FIXME: Keep updates up to date! -->\nI am in the process of implementing breaking changes for a `0.1` version. The major change will be to refactor the various chart constructors to match the plotly express initialization arguments\n\nI plan on reaching >75% test coverage, adding type annotations, and improving the documentation as well\n\nI considered using a builder pattern, but the plotly.express approach can be implemented through an attributes class that accepts unhandled keyword arguments to the plots\n\n```py\n@attr.s(auto..)\nclass ChartSettings(BaseSettings):\n\n    x_label: str\n    y_label: str\n    etc: int\n\n\nclass SomeChart:\n    def __init__(self, **kwargs):\n        self.something = AttrClass(**kwargs)\n```\n\n~~Alternatively, the builder pattern might be useful to create new charts with slightly different views.~~ Would require serializing the class in order to copy rather than modify in memory\n\nCurrently blocked by pending changes to `calcipy` and `calcipy-template` to how doit tasks are configured\n\n## Quick Start\n\n### 1. Install\n\nWith Poetry install `dash_charts` with: `poetry add dash_charts --git https://github.com/KyleKing/dash_charts.git#main`\n\n### 2. Example Code\n\n<!-- CODE:tests/examples/readme.py -->\n\n```py\n"""Example Dash Application."""\n\nfrom typing import Optional\n\nimport dash\nfrom dash import html\nimport plotly.express as px\nfrom box import Box\nfrom implements import implements\n\nfrom dash_charts.pareto_chart import ParetoChart\nfrom dash_charts.utils_app import AppBase, AppInterface\nfrom dash_charts.utils_fig import min_graph\n\n# FIXME: the change to use Box/_ID needs to be implemented in the examples. This is causing failures in the test cases\n#   Still pending if this is the preferred approach\n_ID = Box({\n    \'chart\': \'pareto\',\n})\n"""Default App IDs."""\n\n\n@implements(AppInterface)\nclass ParetoDemo(AppBase):\n    """Example creating a simple Pareto chart."""\n\n    def __init__(self, app: Optional[dash.Dash] = None) -> None:\n        """Initialize app and initial data members. Should be inherited in child class and called with super().\n\n        Args:\n            app: Dash instance. If None, will create standalone app. Otherwise, will be part of existing app\n\n        """\n        self.name = \'Car Share Pareto Demo\'\n        self.data_raw = None\n        self.chart_main = None\n        self._id = _ID\n\n        super().__init__(app=app)\n\n    def generate_data(self) -> None:\n        """Format the car share data from plotly express for the Pareto. Called by parent class."""\n        self.data_raw = (px.data.carshare()\n                         .rename(columns={\'peak_hour\': \'category\', \'car_hours\': \'value\'}))\n        self.data_raw[\'category\'] = [f\'H:{cat:02}\' for cat in self.data_raw[\'category\']]\n\n    def create_elements(self) -> None:\n        """Initialize the charts, tables, and other Dash elements."""\n        self.chart_main = ParetoChart(title=\'Car Share Pareto\', xlabel=\'Peak Hours\', ylabel=\'Car Hours\')\n\n    def return_layout(self) -> dict:\n        """Return Dash application layout.\n\n        Returns:\n            dict: Dash HTML object\n\n        """\n        return html.Div([\n            html.Div([min_graph(\n                id=self._il[self._id.chart],\n                figure=self.chart_main.create_figure(df_raw=self.data_raw),\n            )]),\n        ])\n\n    def create_callbacks(self) -> None:\n        """Register the callbacks."""\n        pass  # Override base class. Not necessary for this example\n\n\nif __name__ == \'__main__\':\n    app = ParetoDemo()\n    app.create()\n    app.run(debug=True)\nelse:\n    app = instance()\n    app.create()\n    FLASK_HANDLE = app.get_server()\n\n```\n\n<!-- /CODE:tests/examples/readme.py -->\n\n### 3. Resulting Pareto Chart\n\n![MinimumExampleScreenCapture](https://raw.githubusercontent.com/KyleKing/dash_charts/main/.images/pareto_readme.png)\n\n### 4. Additional Notes\n\n<!-- FIXME: Document use of the `calcipy_template` instead of manual configuration -->\n\nTO get the most out of the tools, you may want to add `calcipy` and add the below snippets to your `pyproject.toml` file\n\n```toml\n[tool.poetry.dev-dependencies.dash]\nextras = [ "testing",]\nversion = "*, ^1.19"\n\n[tool.poetry.dependencies.calcipy]\ngit = "https://github.com/kyleking/calcipy.git"\nbranch = "dev/development"\nrev = "01635ea"  # Always pin to a commit\ndevelop = true  # Optional: will reinstall each time\n\n[tool.poetry.dev-dependencies.calcipy]\ngit = "https://github.com/kyleking/calcipy.git"\nbranch = "dev/development"\nextras = [ "dev", "lint", "test", "commitizen_legacy"]\n```\n\nYou will also want at minimum the [dodo.py](https://raw.githubusercontent.com/KyleKing/dash_charts/blob/main/dodo.py) and folder structure from [dash_charts](https://github.com/KyleKing/dash_charts)\n\n## Design Principles\n\nI wanted to show that an object oriented programming approach to Dash projects could be really powerful for improving code reuse and improving readability in large projects. Below are a couple of design principles to keep that I try to keep in mind when implementing this package.\n\n- Try to encapsulate all application logic (callbacks, ids, etc.) in application classes or modules\n- For components that are reused between application, create a `module`. The layout, state, and callbacks can all be delegated to the module rather than cluttering the main application\'s class and allowing for better code reuse\n- Move any methods that do not require state (`self`) into standalone functions\n- If you can separate Dash (ids/layout/Callbacks) and Plotly (figures/tables) code, you can generate static HTML or PNGs more easily. Sometimes static output is useful when programmatically generating views rather than tweaking the interactive inputs each time.\n\nOverall, I find that this package really improves my Dash code and I hope others find it useful. Please feel free to submit PRs or open issues if you have any input. See [./DESIGN.md](https://github.com/KyleKing/dash_charts/blob/main/DESIGN.md) for additional information on design decisions and package architecture\n\n## Local Development\n\nInitial commands to clone code from Github, create Python virtual environment, and run first example application\n\n```sh\ngit clone https://github.com/KyleKing/dash_charts.git\ncd dash_charts\npoetry install\npoetry shell\npython tests/examples/ex_app_px.py\n```\n\nOther useful scripts for testing, documentation, and more:\n\n```sh\npoetry run ptw -- -m "not INTERACTIVE"\npoetry run doit run test\npoetry run doit\n```\n\n## Example Charts and Tables\n\nBelow are screenshots and links to the example code for each chart or table included in dash_charts\n\n### Pareto Chart\n\nCreate Pareto charts in Dash. Handles ordering the category, calculating the cumulative percentage, and configuring both YAxis.\n\nSee sample code in [tests/examples/ex_pareto_chart.py](https://www.github.com/KyleKing/dash_charts/blob/main/tests/examples/ex_pareto_chart.py). Screenshot below:\n\n![ex_pareto_chart.png](https://raw.githubusercontent.com/KyleKing/dash_charts/main/.images/ex_pareto_chart.png)\n\n### Gantt Chart\n\nCreate simple Gantt charts in Dash. Doesn\'t support more advanced features such as dependencies and resource assignment, but effectively shows tasks, progress, and projects in a clear way. You can toggle the different projects on/off and use zoom like a normal Plotly chart.\n\nSee sample code in [tests/examples/ex_gantt_chart.py](https://www.github.com/KyleKing/dash_charts/blob/main/tests/examples/ex_gantt_chart.py). Screenshot below:\n\n![ex_gantt_chart.png](https://raw.githubusercontent.com/KyleKing/dash_charts/main/.images/ex_gantt_chart.png)\n\n### Time Vis Chart\n\nCreate a TimeVis chart to plot general time visualizations (based on [daattali/timevis](https://github.com/daattali/timevis) from R). This chart differs from a Gantt chart by showing events that repeat in the same row.\n\nSee sample code in [tests/examples/ex_time_vis_chart.py](https://www.github.com/KyleKing/dash_charts/blob/main/tests/examples/ex_time_vis_chart.py). Screenshot below:\n\n![ex_time_vis_chart.png](https://raw.githubusercontent.com/KyleKing/dash_charts/main/.images/ex_time_vis_chart.png)\n\n### Rolling Mean and STD Chart\n\nEasily chart the rolling mean and standard deviation for a given scatter data set. Demonstrates using chart annotations.\n\nSee sample code in [tests/examples/ex_rolling_chart.py](https://www.github.com/KyleKing/dash_charts/blob/main/tests/examples/ex_rolling_chart.py). Screenshot below:\n\n![ex_rolling_chart.png](https://raw.githubusercontent.com/KyleKing/dash_charts/main/.images/ex_rolling_chart.png)\n\n### Fitted Chart\n\nFit arbitrary scatter data with a specified equation. On hover will show the rounded fit parameters. Useful for experimenting with new data.\n\nSee sample code in [tests/examples/ex_fitted_chart.py](https://www.github.com/KyleKing/dash_charts/blob/main/tests/examples/ex_fitted_chart.py). Screenshot below:\n\n![ex_fitted_chart.png](https://raw.githubusercontent.com/KyleKing/dash_charts/main/.images/ex_fitted_chart.png)\n\n### Real Time SQL Demo\n\nExample using a Rolling Chart to show real time data from a SQLite database. For the demo, a background thread populates the database as the Dash app runs in a separate thread.\n\nSee sample code in [tests/examples/ex_sqlite_realtime.py](https://www.github.com/KyleKing/dash_charts/blob/main/tests/examples/ex_sqlite_realtime.py). Screenshot below:\n\n![ex_sqlite_realtime.gif](https://raw.githubusercontent.com/KyleKing/dash_charts/main/.images/ex_sqlite_realtime.gif)\n\n### Coordinate Chart\n\nChart a discrete data set on a 2D plane with color for intensity. Below examples show how to use the `YearGrid()`, `MonthGrid()`, and `CircleGrid()` classes\n\nSee sample code in [tests/examples/ex_coordinate_chart.py](https://www.github.com/KyleKing/dash_charts/blob/main/tests/examples/ex_coordinate_chart.py). Screenshot below:\n\n![ex_coordinate_chart.png](https://raw.githubusercontent.com/KyleKing/dash_charts/main/.images/ex_coordinate_chart.png)\n\n### Marginal Chart\n\nExample creating a new chart from utils_fig.MarginalChart\n\nSee sample code in [tests/examples/ex_marginal_chart.py](https://www.github.com/KyleKing/dash_charts/blob/main/tests/examples/ex_marginal_chart.py). Screenshot below:\n\n![ex_marginal_chart.png](https://raw.githubusercontent.com/KyleKing/dash_charts/main/.images/ex_marginal_chart.png)\n\n### Data Table Module\n\nDisplay Dash data table from dataframe\n\nSee sample code in [tests/examples/ex_datatable.py](https://www.github.com/KyleKing/dash_charts/blob/main/tests/examples/ex_datatable.py). Screenshot below:\n\n![ex_datatable.png](https://raw.githubusercontent.com/KyleKing/dash_charts/main/.images/ex_datatable.png)\n\n### Upload Module\n\nUpload module for user-selected CSV, JSON, or SQLite files.\n\nSee sample code in [tests/examples/ex_upload.py](https://ra/www.gubusem/KyleKing/dash_charts/blob/main/tests/examples/ex_upload.py). Screenshot below:\n\n![ex_upload.png](https://raw.githubusercontent.com/KyleKing/dash_charts/main/.images/ex_upload.png)\n\n## Applications\n\nEvery app derives from `AppBase()` so that each tab or page can be run independently or part of a more complicated application\n\n### Tabbed Application\n\nUse the `AppWithTabs()` base class to quickly build applications with tabbed navigation. You can set tabs to the top/bottom/left/right, to be compact or not, etc.\n\nSee sample code in [tests/examples/ex_tabs.py](https://ra/www.gubusem/KyleKing/dash_charts/blob/main/tests/examples/ex_tabs.py). Screenshot below:\n\n![ex_tabs.png](https://raw.githubusercontent.com/KyleKing/dash_charts/main/.images/ex_tabs.png)\n\n### Multi-Page Application\n\nUse the `AppMultiPage()` base class to quickly build applications with tabbed navigation. You can set tabs to the top/bottom/left/right, to be compact or not, etc.\n\nSee sample code in [tests/examples/ex_multi_page.py](https://www.github.com/KyleKing/dash_charts/blob/main/tests/examples/ex_multi_page.py). Screenshot below:\n\n![ex_multi_page.png](https://raw.githubusercontent.com/KyleKing/dash_charts/main/.images/ex_multi_page.png)\n\n### PX Generic Application\n\nTo facilitate general data analysis from a JSON, CSV, or SQLite file, dash_charts includes apps for each px plot type and a tabbed app to bring them all together.\n\nSee sample code in [tests/examples/ex_app_px.py](https://ra/www.gubusem/KyleKing/dash_charts/blob/main/tests/examples/ex_app_px.py). Screenshot below:\n\n![ex_app_px.png](https://raw.githubusercontent.com/KyleKing/dash_charts/main/.images/ex_app_px.png)\n\n## Other\n\nOther notable components in `dash_charts`\n\n### Cache\n\nUtilities for utilizing a local cache file system for storing JSON data. Particularly useful if scraping or making many API calls to cache the responses locally.\n\nSee sample code in the relevant test file [tests/test_utils_json_cache.py](https:/raw./usercontentgithub.com/KyleKing/dash_charts/blob/main/tests/test_utils_json_cache.py)\n\n### Static HTML Generation\n\nutilities for creating static HTML output with all of the CustomCharts above or user-created. Also includes utilities for writing tables, code, markdown, and more to be added.\n\nSee sample code in [tests/examples/ex_utils_static.py](https://www.github.com/KyleKing/dash_charts/blob/main/tests/examples/ex_utils_static.py). Screenshot below:\n\n![ex_utils_static.png](https://raw.githubusercontent.com/KyleKing/dash_charts/main/.images/ex_utils_static.png)\n\n## Coverage\n\nLatest coverage table\n\n<!-- COVERAGE -->\n\n| File | Statements | Missing | Excluded | Coverage |\n| --: | --: | --: | --: | --: |\n| `dash_charts/__init__.py` | 2 | 0 | 0 | 100.0% |\n| `dash_charts/app_px.py` | 130 | 11 | 0 | 91.5% |\n| `dash_charts/components.py` | 13 | 0 | 0 | 100.0% |\n| `dash_charts/coordinate_chart.py` | 102 | 1 | 6 | 99.0% |\n| `dash_charts/custom_colorscales.py` | 3 | 0 | 0 | 100.0% |\n| `dash_charts/datatable.py` | 79 | 25 | 0 | 68.4% |\n| `dash_charts/equations.py` | 11 | 0 | 0 | 100.0% |\n| `dash_charts/gantt_chart.py` | 54 | 0 | 0 | 100.0% |\n| `dash_charts/modules_datatable.py` | 101 | 11 | 0 | 89.1% |\n| `dash_charts/modules_upload.py` | 132 | 60 | 0 | 54.5% |\n| `dash_charts/pareto_chart.py` | 43 | 0 | 2 | 100.0% |\n| `dash_charts/scatter_line_charts.py` | 45 | 0 | 3 | 100.0% |\n| `dash_charts/time_vis_chart.py` | 61 | 0 | 0 | 100.0% |\n| `dash_charts/utils_app.py` | 98 | 14 | 6 | 85.7% |\n| `dash_charts/utils_app_modules.py` | 26 | 3 | 4 | 88.5% |\n| `dash_charts/utils_app_with_navigation.py` | 119 | 9 | 6 | 92.4% |\n| `dash_charts/utils_callbacks.py` | 34 | 6 | 0 | 82.4% |\n| `dash_charts/utils_data.py` | 63 | 1 | 0 | 98.4% |\n| `dash_charts/utils_dataset.py` | 76 | 43 | 0 | 43.4% |\n| `dash_charts/utils_fig.py` | 77 | 2 | 4 | 97.4% |\n| `dash_charts/utils_helpers.py` | 17 | 8 | 7 | 52.9% |\n| `dash_charts/utils_json_cache.py` | 51 | 10 | 0 | 80.4% |\n| `dash_charts/utils_static.py` | 111 | 5 | 0 | 95.5% |\n| `dash_charts/utils_static_toc.py` | 22 | 1 | 0 | 95.5% |\n\nGenerated on: 2020-11-08T22:46:27.420973\n\n<!-- /COVERAGE -->\n\n## External Links\n\n[![codebeat badge](https://codebeat.co/badges/9d2b9a53-9203-4467-8cde-68f24e930389)](https://codebeat.co/projects/github-com-kyleking-dash_charts-main) [![Documentation Site Online Status Badge](https://img.shields.io/website?label=Doc%20Site&up_message=up&url=https%3A%2F%2Fgithub.com%2FKyleKing%2Fdash_charts%2F)](https://kyleking.me/dash_charts/) ![Commits Since last Release](https://img.shields.io/github/commits-since/KyleKing/dash_charts/latest) ![Last Commit Badge](https://img.shields.io/github/last-commit/kyleking/dash_charts)\n\nUseful Dash reference links\n\n- Official [Dash Docs](https://dash.plot.ly) / [Plotly Docs](https://plot.ly/python/)\n- Example Apps\n    - Pretty annotations from [Market Ahead](https://www.marketahead.com/p/FOX), a commercial Dash app\n    - Pleasant dark app from Dash-Bio, [Circos](https://github.com/plotly/dash-bio/blob/main/tests/dashbio_demos/app_circos.py)\n    - All [Dash Sample Apps](https://github.com/plotly/dash-sample-apps/tree/main/apps)\n- Code Conceptual inspiration\n    - [On Tidy data](https://www.jeannicholashould.com/tidy-data-in-python.html)\n    - [Using field for properties in a @DataClass](https://florimond.dev/blog/articles/2018/10/reconciling-dataclasses-and-properties-in-python/)\n- Best Practices\n    - IBM Design Language (note that v2 doesn\'t have documentation on visualization yet)\n        - [Process](https://www.ibm.com/design/v1/language/experience/data-visualization/process/)\n        - [Select Chart](https://www.ibm.com/design/v1/language/experience/data-visualization/chart-models/)\n            - Alt [Chart Decision Diagram](https://www.tatvic.com/blog/7-visualizations-learn-r/)\n        - [Principles (color conventions, etc.)](https://www.ibm.com/design/v1/language/experience/data-visualization/visualization/)\n        - [Color Library and Data Vis Swatches](https://www.ibm.com/design/v1/language/resources/color-library/)\n        - [Interaction](https://www.ibm.com/design/v1/language/experience/data-visualization/interaction/)\n        - [Style](https://www.ibm.com/design/v1/language/experience/data-visualization/style/)\n- Colors\n    - [Friendly Guide to Colors](https://lisacharlotterost.de/2016/04/22/Colors-for-DataVis/)\n    - [Viz-Palette Tool](https://projects.susielu.com/viz-palette)\n    - [AI Color Palette](http://colormind.io/) / [Coolors](https://coolors.co/2b303a-92dce5-eee5e9-7c7c7c-d64933)\n- Reference\n    - [Pandas CheatSheet for Data Manipulation](https://github.com/pandas-dev/pandas/blob/main/doc/cheatsheet/Pandas_Cheat_Sheet.pdf)\n\n<!-- TODO: See https://github.com/KyleKing/calcipy/issues/38 -->\n\n## Installation\n\n1. ...\n2. ...\n\n    ```sh\n    echo Hello World!\n    ```\n\n3. ...\n\n## Usage\n\n<!-- TODO: Show an example (screenshots, terminal recording, etc.) -->\n\nFor more examples, see [Scripts](https://github.com/kyleking/dash_charts/scripts) or [Tests](https://github.com/kyleking/dash_charts/tests)\n\n## Roadmap\n\nSee the `Open Issues` and `Milestones` for current status and [./docs/CODE_TAG_SUMMARY.md](./docs/CODE_TAG_SUMMARY.md) for annotations in the source code.\n\nFor release history, see the [./docs/CHANGELOG.md](./docs/CHANGELOG.md)\n\n## Contributing\n\nSee the Developer Guide, Contribution Guidelines, etc\n\n- [./docs/DEVELOPER_GUIDE.md](./docs/DEVELOPER_GUIDE.md)\n- [./docs/STYLE_GUIDE.md](./docs/STYLE_GUIDE.md)\n- [./docs/CONTRIBUTING.md](./docs/CONTRIBUTING.md)\n- [./docs/CODE_OF_CONDUCT.md](./docs/CODE_OF_CONDUCT.md)\n- [./docs/SECURITY.md](./docs/SECURITY.md)\n\n## License\n\n[LICENSE](https://github.com/kyleking/dash_charts/LICENSE)\n',
    'author': 'Kyle King',
    'author_email': 'dev.act.kyle@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kyleking/dash_charts',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Generic Plotly Express Data Analysis App(s).

Examples: https://www.plotly.express/

Docs: https://www.plotly.express/plotly_express/

# (Currently) Unsupported plotly express types

```py
px.parallel_coordinates(px.data.iris(), color="species_id",
                        dimensions=['sepal_width', 'sepal_length', 'petal_width', 'petal_length'])
px.treemap(px.data.tips(), path=['day', 'time', 'sex'], values='total_bill')
px.sunburst(px.data.tips(), path=['day', 'time', 'sex'], values='total_bill)
```

Other charts that could be useful (but won't work with simple dropdowns)

- scatter_matrix([data_frame, dimensions, …])
    - In a scatter plot matrix (or SPLOM), each row of data_frame is
    - https://plotly.com/python/splom/
- parallel_coordinates([data_frame, …])
    - In a parallel coordinates plot, each row of data_frame is represented
    - https://plotly.com/python-api-reference/generated/plotly.express.parallel_coordinates.html
- parallel_categories([data_frame, …])
    - In a parallel categories (or parallel sets) plot, each row of
    - https://plotly.com/python-api-reference/generated/plotly.express.parallel_categories.html
- density_heatmap([data_frame, x, y, z, …])
    - In a density heatmap, rows of data_frame are grouped together into
    - https://plotly.com/python-api-reference/generated/plotly.express.density_heatmap.html
- imshow(img[, zmin, zmax, origin, …])
    - Display an image, i.e.
    - https://plotly.com/python/imshow/

"""

from collections import OrderedDict

import dash
import dash_bootstrap_components as dbc
import pandas as pd
import plotly.express as px
from dash import html
from implements import implements

from .components import dropdown_group, opts_dd
from .utils_app import AppBase, AppInterface
from .utils_app_with_navigation import FullScreenAppWithTabs
from .utils_callbacks import map_args, map_outputs
from .utils_fig import min_graph

# ======================================================================================================================
# Create classes to manage tabs state. Easy to scale up or down
# >> Demo uses sample data. User could replace with data loaded from a static CSV file, TinyDB, SQLite, etc.


@implements(AppInterface)  # noqa: H601
class TabBase(AppBase):
    """Base tab class with helper methods."""

    external_stylesheets = [dbc.themes.FLATLY]

    # ID Elements for UI
    id_chart: str = 'chart'
    id_func: str = 'func'
    id_template: str = 'template'  # PLANNED: template should be able to be None

    takes_args: bool = True
    """If True, will pass arguments from UI to function."""

    templates: list = [
        'ggplot2', 'seaborn', 'simple_white', 'plotly',
        'plotly_white', 'plotly_dark', 'presentation', 'xgridoff',
        'ygridoff', 'gridon', 'none',
    ]
    """List of templates from: `import plotly.io as pio; pio.templates`"""

    # Must override in child class
    name: str = None
    """Unique tab component name. Must be overridden in child class."""
    data: pd.DataFrame = None
    """Dataframe. Must be overridden in child class."""
    func_map: OrderedDict = None
    """Map of functions to keywords. Must be overridden in child class."""

    # PLANNED: below items should be able to be None
    dims: tuple = ()
    """Keyword from function for dropdowns with column names as options. Must be overridden in child class."""
    dims_dict: OrderedDict = OrderedDict([])
    """OrderedDict of keyword from function to allowed values. Must be overridden in child class."""
    default_dim_name = {}
    """Lookup for dim:column name to use as default in dropdown."""

    def initialization(self) -> None:
        """Initialize ids with `self.register_uniq_ids([...])` and other one-time actions."""
        super().initialization()

        # Register the the unique element IDs
        self.input_ids = [self.id_func, self.id_template] + [*self.dims] + [*self.dims_dict.keys()]
        self.register_uniq_ids([self.id_chart] + self.input_ids)

        # Configure the options for the various dropdowns
        self.col_opts = [] if self.data is None else tuple(opts_dd(_c, _c) for _c in self.data.columns)
        self.func_opts = tuple(opts_dd(lbl, lbl) for lbl in self.func_map.keys())
        self.t_opts = tuple(opts_dd(template, template) for template in self.templates)

    def create_elements(self) -> None:
        """Initialize the charts, tables, and other Dash elements."""
        ...

    def verify_types_for_layout(self):
        """Verify data types of data members necessary for the layout of this tab.

        Raises:
            RuntimeError: if any relevant data members are of the wrong type

        """
        errors = []
        if not isinstance(self.name, str):
            errors.append(f'Expected self.name="{self.name}" to be str')
        if not isinstance(self.dims, tuple):
            errors.append(f'Expected self.dims="{self.dims}" to be tuple')
        if not isinstance(self.dims_dict, OrderedDict):
            errors.append(f'Expected self.dims_dict="{self.dims_dict}" to be OrderedDict')
        if errors:
            formatted_errors = '\n' + '\n'.join(errors)
            raise RuntimeError(f'Found errors in data members:{formatted_errors}')

    def verify_types_for_callbacks(self):
        """Verify data types of data members necessary for the callbacks of this tab.

        Raises:
            RuntimeError: if any relevant data members are of the wrong type

        """
        errors = []
        if not isinstance(self.takes_args, bool):
            errors.append(f'Expected self.takes_args="{self.takes_args}" to be bool')
        if not (isinstance(self.data, pd.DataFrame) or self.data is None):
            errors.append(f'Expected self.data="{self.data}" to be pd.DataFrame or None')
        if not isinstance(self.func_map, OrderedDict):
            errors.append(f'Expected self.func_map="{self.func_map}" to be OrderedDict')
        if errors:
            formatted_errors = '\n' + '\n'.join(errors)
            raise RuntimeError(f'Found errors in data members:{formatted_errors}')

    def return_layout(self) -> dict:
        """Return Dash application layout.

        Returns:
            dict: Dash HTML object

        """
        self.verify_types_for_layout()

        return html.Div(
            [  # noqa: ECE001
                html.Div(
                    [
                        dropdown_group(
                            'Plot Type:', self._il[self.id_func],
                            self.func_opts, value=self.func_opts[0]['label'],
                        ),
                        dropdown_group(
                            'Template:', self._il[self.id_template],
                            self.t_opts, value=self.t_opts[0]['label'],
                        ),
                    ] + [
                        dropdown_group(
                            f'{dim}:', self._il[dim], self.col_opts,
                            value=self.default_dim_name.get(dim, None),
                        )
                        for dim in self.dims
                    ] + [
                        dropdown_group(f'{dim}:', self._il[dim], [opts_dd(item, item) for item in items])
                        for dim, items in self.dims_dict.items()
                    ], style={'width': '25%', 'float': 'left'},
                ),
                min_graph(id=self._il[self.id_chart], style={'width': '75%', 'display': 'inline-block'}),
            ], style={'padding': '15px'},
        )

    def create_callbacks(self) -> None:
        """Register callbacks necessary for this tab."""
        self.verify_types_for_callbacks()

        self.register_update_chart()

    def register_update_chart(self):   # noqa: CCR001
        """Register the update_chart callback."""
        outputs = [(self.id_chart, 'figure')]
        inputs = [(_id, 'value') for _id in self.input_ids]
        states = ()

        @self.callback(outputs, inputs, states)
        def update_chart(*raw_args):
            a_in, _a_states = map_args(raw_args, inputs, states)
            name_func = a_in[self.id_func]['value']

            properties = [trigger['prop_id'] for trigger in dash.callback_context.triggered]
            new_chart = {}
            # If event is not a tab change, return the updated chart
            if 'tabs-select.value' not in properties:  # FIXME: replace tabs-select with actual keyname (?)
                if self.takes_args:
                    # Parse the arguments to generate a new plot
                    kwargs = {key: a_in[key]['value'] for key in self.input_ids[1:]}
                    new_chart = self.func_map[name_func](self.data, height=650, **kwargs)
                else:
                    new_chart = self.func_map[name_func]()
            # Example Mapping Output. Alternatively, just: `return [new_chart]`
            return map_outputs(outputs, [(self.id_chart, 'figure', new_chart)])


@implements(AppInterface)  # noqa: H601
class TabTip(TabBase):
    """TabTip properties."""

    name = 'Tip Data'
    data = px.data.tips()
    func_map = OrderedDict([
        ('scatter', px.scatter),
        ('density_contour', px.density_contour),
    ])
    dims = ('x', 'y', 'color', 'facet_row', 'facet_col')
    dims_dict = OrderedDict([
        ('marginal_x', ('histogram', 'rag', 'violin', 'box')),
        ('marginal_y', ('histogram', 'rag', 'violin', 'box')),
        ('trendline', ('ols', 'lowess')),
    ])
    default_dim_name = {
        'x': 'total_bill',
        'y': 'tip',
        'color': 'smoker',
    }


@implements(AppInterface)  # noqa: H601
class TabIris(TabBase):
    """TabIris properties."""

    name = 'Iris Data'
    data = px.data.iris()
    func_map = OrderedDict([
        ('histogram', px.histogram),
        ('bar', px.bar),
        ('strip', px.strip),
        ('box', px.box),
        ('violin', px.violin),
    ])
    dims = ('x', 'y', 'color', 'facet_row', 'facet_col')
    default_dim_name = {
        'x': 'sepal_width',
        'color': 'species',
    }


@implements(AppInterface)  # noqa: H601
class TabGapminder(TabBase):
    """TabGapminder properties."""

    name = 'Gapminder Data'
    data = px.data.gapminder()
    func_map = OrderedDict([
        ('area', px.area),
        ('line', px.line),
    ])
    dims = ('x', 'y', 'color', 'line_group', 'facet_row', 'facet_col')
    default_dim_name = {
        'x': 'year',
        'y': 'pop',
        'color': 'continent',
        'line_group': 'country',
    }


@implements(AppInterface)  # noqa: H601
class TabTernary(TabBase):
    """TabTernary properties."""

    name = 'Ternary'
    data = px.data.election()
    func_map = OrderedDict([
        ('scatter_ternary', px.scatter_ternary),
        ('line_ternary', px.line_ternary),
    ])
    dims = ('a', 'b', 'c', 'color', 'hover_name')  # size - only for scatter
    default_dim_name = {
        'a': 'Joly',
        'b': 'Coderre',
        'c': 'Bergeron',
        'color': 'winner',
        'hover_name': 'district',
    }


@implements(AppInterface)  # noqa: H601
class TabWind(TabBase):
    """TabWind properties."""

    name = 'Wind'
    data = px.data.wind()
    func_map = OrderedDict([
        ('scatter_polar', px.scatter_polar),
        ('line_polar', px.line_polar),  # (line_close=True)
        ('bar_polar', px.bar_polar),
    ])
    dims = ('r', 'theta', 'color')
    default_dim_name = {
        'r': 'frequency',
        'theta': 'direction',
        'color': 'strength',
        'symbol': 'strength',
    }


@implements(AppInterface)  # noqa: H601
class TabColor(TabBase):
    """TabColor properties."""

    name = 'Color Swatches'
    takes_args = False
    func_map = OrderedDict([
        ('colors.qualitative', px.colors.qualitative.swatches),
        ('colors.sequential', px.colors.sequential.swatches),
        ('colors.diverging', px.colors.diverging.swatches),
        ('colors.cyclical', px.colors.cyclical.swatches),
        ('colors.colorbrewer', px.colors.colorbrewer.swatches),
        ('colors.cmocean', px.colors.cmocean.swatches),
        ('colors.carto', px.colors.carto.swatches),
    ])
    default_dim_name = {
        'x': 'sepal_width',
        'y': 'sepal_length',
    }

# ======================================================================================================================
# Create class for application to control manage variable scopes


class InteractivePXApp(FullScreenAppWithTabs):  # noqa: H601
    """Plotly Express Demo application."""

    name = 'TabAppDemo'

    tabs_location = 'right'
    """Tab orientation setting. One of `(left, top, bottom, right)`."""

    tabs_margin = '175px'
    """Adjust this setting based on the width or height of the tabs to prevent the content from overlapping the tabs."""

    tabs_compact = False
    """Boolean setting to toggle between a padded tab layout if False and a minimal compact version if True."""

    def define_nav_elements(self):
        """Return list of initialized tabs.

        Returns:
            list: each item is an initialized tab (ex `[AppBase(self.app)]` in the order each tab is rendered

        """
        return [
            TabTip(app=self.app),
            TabIris(app=self.app),
            TabGapminder(app=self.app),
            TabTernary(app=self.app),
            TabWind(app=self.app),
            TabColor(app=self.app),
        ]

    def return_layout(self) -> dict:
        """Return Dash application layout.

        Returns:
            dict: Dash HTML object

        """
        return html.Div([
            html.H1('Dash/Plotly Express Data Exploration Demo', style={'padding': '15px 0 0 15px'}),
            super().return_layout(),
        ])



"""Helpers for building Dash applications."""

import argparse
import time
from datetime import datetime
from typing import Optional

from beartype import beartype

# ----------------------------------------------------------------------------------------------------------------------
# General Debug


@beartype
def debug_time(message: str, last: Optional[datetime] = None) -> datetime:
    """Debug timing issues.

    Args:
        message: string message to print
        last: last timestamp

    Returns:
        timestamp: the current timestamp to calculate the next delta

    """
    if last is None:
        last = time.time()
    now = time.time()
    delta = now - last
    if delta > 0.5:
        print(message, delta)  # noqa: T001
    return now


# ----------------------------------------------------------------------------------------------------------------------
# Dash Helpers


def parse_dash_cli_args():  # pragma: no cover
    """Configure the CLI options for Dash applications.

    Returns:
        dict: keyword arguments for Dash

    """
    parser = argparse.ArgumentParser(description='Process Dash Parameters.')
    parser.add_argument(
        '--port', type=int, default=8050,
        help='Pass port number to Dash server. Default is 8050',
    )
    parser.add_argument(
        '--nodebug', action='store_true', default=False,
        help='If set, will disable debug mode. Default is to set `debug=True`',
    )
    args = parser.parse_args()
    return {'port': args.port, 'debug': not args.nodebug}


# ----------------------------------------------------------------------------------------------------------------------
# Functional Programming


def graph_return(resp, keys):
    """Based on concepts of GraphQL, return specified subset of response.

    Args:
        resp: dictionary with values from function
        keys: list of keynames from the resp dictionary

    Returns:
        the `resp` dictionary with only the keys specified in the `keys` list

    Raises:
        RuntimeError: if `keys` is not a list or tuple

    """
    if not (len(keys) and isinstance(keys, (list, tuple))):
        raise RuntimeError(f'Expected list of keys for: `{resp.items()}`, but received `{keys}`')
    ordered_responses = [resp.get(key, None) for key in keys]
    return ordered_responses if len(ordered_responses) > 1 else ordered_responses[0]
